from ..exceptions import (OutOfAdditionDict, OutOfList, OutOfRange, OutOfDict,
                          InvalidStrLength, InvalidNumFormat, NotNumericValue)


class FormatChecker:
    def __init__(self, cell, dics, input_type, row_type=None):
        self._cell = cell
        self._dics = dics

        self.row_type = row_type
        self.input_type = input_type

        self.dic = cell.attrib.get('dic')
        self.format = cell.attrib.get('format')
        self.vld_type = cell.attrib.get('vldType')
        self.vld_param = cell.attrib.get('vld')

        self.format_funcs_map = {'N': self._is_num, 'C': self._is_chars}

    def __repr__(self):
        return ('<FormatChecker row={row_type} input={input_type} '
                'format={format} vld_type={vld_type} '
                'vld_param={vld_param}>').format(**self.__dict__)

    @classmethod
    def _is_num(self, value, limits):
        '''Проверка длины целой и дробной частей числового значения поля'''
        try:
            float(value)
        except ValueError:
            raise NotNumericValue()

        value_parts = tuple(len(n) for n in value.split('.'))
        if len(value_parts) == 1:
            i_part_len, f_part_len = value_parts[0], 0
        else:
            i_part_len, f_part_len = value_parts[0], value_parts[1]
        i_part_lim, f_part_lim = (int(n) for n in limits.split(','))

        if not (i_part_len <= i_part_lim and f_part_len <= f_part_lim):
            raise InvalidNumFormat()

    @classmethod
    def _is_chars(self, value, limit):
        '''Проверка длины символьного значения поля'''
        if not len(value) <= limit:
            raise InvalidStrLength()

    def check(self, cell, errors_list):
        '''Метод вызова проверки формата значения'''
        self._check_format(cell)
        self._check_value(cell)

    def _check_format(self, cell):
        '''Разбор "формулы" проверки формата. Вызов метода проверки'''
        alias, args = self.format.strip(' )').split('(')
        format_check_func = self.format_funcs_map[alias]
        format_check_func(cell, args)

    def _check_value(self, cell):
        '''Определение и вызов метода проверки по типу'''
        if self.vld_type == '1':
            self.__check_value_dic(cell)
        elif self.vld_type == '2':
            self.__check_value_range(cell)
        elif self.vld_type == '3':
            self.__check_value_list(cell)
        elif self.vld_type == '4':
            self.__check_value_dic_add(cell)
        elif self.vld_type == '5':
            self.__check_value_dic_coord(cell)

    def __check_value_dic(self, cell):
        '''Проверка на вхождение в справочник'''
        if cell not in self._dics[self.dic]:
            raise OutOfDict()

    def __check_value_range(self, cell):
        '''Проверка на вхождение в диапазон'''
        cell = float(cell)
        start, end = (int(n) for n in self.vld_param.split('-'))
        if not (cell >= start and cell <= end):
            raise OutOfRange()

    def __check_value_list(self, cell):
        '''Проверка на вхождение в список'''
        if cell not in self.vld_param.split(','):
            raise OutOfList()

    def __check_value_dic_add(self, cell):
        '''Проверка на вхождение в справочник приложение'''
        if cell not in self._dics[self.vld_param]:
            raise OutOfAdditionDict()

    def __check_value_coord(self, cell):
        return  # пока не ясно как это првоеряется
        attr, coords = self.vld_param.split('=#')
        s_idx, r_idx, c_idx = coords.split(',')



from lxml import etree
from .report import Report
from .validator.schema import Schema


def _get_xml_etree(source):
    if isinstance(source, str):
        return etree.parse(source)
    return source


def parse_report(source):
    xml_etree = _get_xml_etree(source)
    return Report(xml_etree)


def parse_schema(source):
    xml_etree = _get_xml_etree(source)
    return Schema(xml_etree)



import re
from ..exceptions import PeriodExprError

in_pattern = re.compile(r'^\(&NP\s?in\s?\(([\d, ]+)\)\)$', re.I)
sp_pattern = re.compile(r'^\(&NP\s?([><=]+)\s?(\d+)\)$', re.I)
cp_pattern = re.compile(r'^\(&NP\s?([><=]+)\s?(\d+)\s?(and|or)\s?'
                        r'&NP\s?([><=]+)\s?(\d+)\)$', re.I)


class PeriodClause:
    def __init__(self, control, _id):
        self.id = _id
        self.period_clause = control.attrib.get('periodClause', '').strip()

    def __repr__(self):
        return '<PeriodClause clause={period_clause}>'.format(**self.__dict__)

    def check(self, report):
        '''Метод вызова проверки периода контроля'''
        if not self.period_clause:
            return True

        if 'in' in self.period_clause:
            return self._check_in(report)
        elif 'or' in self.period_clause or 'and' in self.period_clause:
            return self._check_complex(report)
        else:
            return self._check_simple(report)

    def _eval_regex(self, pattern, string):
        '''Разбор формулы проверки периода с помощью регулярки'''
        result = pattern.match(string)
        if result is None:
            raise PeriodExprError(self.id)
        return result

    def _check_in(self, report):
        '''Проверка на вхождение в список'''
        clause_parts = self._eval_regex(in_pattern, self.period_clause)

        clause = '{0} in ({1},)'.format(
            report.period_code, clause_parts.group(1))
        return eval(clause)

    def _check_complex(self, report):
        '''Проверка сложного логического условия'''
        clause_parts = self._eval_regex(cp_pattern, self.period_clause)

        l_op = '==' if clause_parts.group(1) == '=' else clause_parts.group(1)
        r_op = '==' if clause_parts.group(4) == '=' else clause_parts.group(4)

        clause = '{0} {1} {3} {4} {0} {2} {5}'.format(
            report.period_code, l_op, r_op, *clause_parts.group(2, 3, 5))
        return eval(clause)

    def _check_simple(self, report):
        '''Проверка простого логического условия'''
        clause_parts = self._eval_regex(sp_pattern, self.period_clause)

        op = '==' if clause_parts.group(1) == '=' else clause_parts.group(1)

        clause = '{0} {1} {2}'.format(
            report.period_code, op, clause_parts.group(2))
        return eval(clause)



from ..controls.period import PeriodClause
from ..controls import parser as control_parser
from ..exceptions import (PeriodCheckFail, ConditionExprError, RuleExprError,
                          ConditionCheckFail, RuleCheckFail, PrevPeriodNotImpl)


class ControlChecker:
    def __init__(self, control):
        self._control = control

        self.id = control.attrib['id']
        self.name = control.attrib['name']
        self.rule = control.attrib['rule'].strip()
        self.condition = control.attrib['condition'].strip()

        self.tip = control.attrib.get('tip', '1')
        self.fault = float(control.attrib.get('fault', '0'))
        self.precision = int(control.attrib.get('precision', '2'))

        self.period = PeriodClause(control, self.id)

    def __repr__(self):
        return ('<ControlChecker id={id} name={name} rule={rule} '
                'condition={condition} tip={tip} fault={fault} '
                'period={period} precision={precision}>').format(
                    **self.__dict__)

    def check(self, report, errors_list) -> None:
        '''Метод вызова проверки контроля'''
        try:
            self._check_period(report)
            self._check_condition(report)
            self._check_rule(report)
        except (PeriodCheckFail, ConditionCheckFail):
            pass
        except RuleCheckFail as ex:
            errors_list.extend(self._fmt_errors(ex.msg))
        except (ConditionExprError, RuleExprError, PrevPeriodNotImpl) as ex:
            errors_list.append(ex.msg)

    def _fmt_errors(self, errors):
        '''Форматирование сообщения о непройденном контроле'''
        template = '{} {}; слева {} {} справа {} разница {}; обязательность {}'
        for err in errors:
            yield template.format(self.id,
                                  self.name,
                                  err['left'],
                                  err['operator'],
                                  err['right'],
                                  err['delta'],
                                  'да' if self.tip == '1' else 'нет')

    def _check_period(self, report):
        '''Проверка соответствия периода контроля периоду в отчёте'''
        if not self.period.check(report):
            raise PeriodCheckFail()

    def __check_control(self, evaluator, report, rule=False):
        '''Выполнение проверки. Возвращает список проваленых проверок'''
        flatten = []
        for result in evaluator.check(report,
                                      fault=self.fault if rule else -1,
                                      precision=self.precision):
            flatten.extend(result.controls)
        return flatten

    def _check_condition(self, report):
        '''Проверка условия для выполнения контроля'''
        if self.condition and not self._is_previous_period(self.condition):
            condition = control_parser.parse(self.condition)
            if condition is None:
                raise ConditionExprError(self.id)

            fail_checks = self.__check_control(condition, report)
            if fail_checks:
                raise ConditionCheckFail()

    def _check_rule(self, report):
        '''Проверка правила контроля'''
        if self.rule and not self._is_previous_period(self.rule):
            rule = control_parser.parse(self.rule)
            if rule is None:
                raise RuleExprError(self.id)

            fail_checks = self.__check_control(rule, report, rule=True)
            if fail_checks:
                raise RuleCheckFail(fail_checks)

    def _is_previous_period(self, formula):
        '''Проверка наличия в формуле элемента в двух фигурных скобках,
           что говорит о том, что значение берётся за прошлый период.
           Такой функционал пока неизвестно когда получится реализовать
        '''
        if '{{' in formula:
            raise PrevPeriodNotImpl(self.id)
